#lang scheme

(require scheme/runtime-path
         "test.ss")

;; A board is a vector-of-vector-of-cell
;; A cell is either
;;  'out --- not a place you can walk on
;;  'space --- a place you can walk on
;;  room --- part of a room
;;  door --- a door to the corresponding room
;; A room is one of 'C 'B 'L 'S 'O 'H 'K 'D 'U
;; A door is one of 'c 'b 'l 's 'o 'h 'k 'd 'u
;; A mode is one of 'V 'S
;; A dimen is an exact integer in [0,24]
;; A pos is (list dimen dimen)
;; A room-pos is (list room mode)
;; A dest is one of pos room-pos
;; A state is a list-of-dest

(provide board-ref
         space?
         solid?
         door? door->room
         room?
         find-doors)

;; Define path relative to source file:
(define-runtime-path clue-board "clue-board.txt")

(define board
  (with-input-from-file clue-board
    (lambda ()
      (list->vector
       (for/list ([row (in-range 25)])
         (begin0
           (list->vector
            (for/list ([col (in-range 25)])
              (let ([c (read-char)])
                (cond
                  [(or (eq? c #\+) (char-numeric? c)) 'out]
                  [(eq? c '#\-) 'space]
                  [(or (char-upper-case? c)
                       (char-lower-case? c))
                   (string->symbol (string c))]
                  [else (error 'parse-board "huh?: ~e" c)]))))
           (read-byte)))))))

;; board-ref : int int -> cell
(define (board-ref row col)
  (if (and (<= 0 row 24)
           (<= 0 col 24))      
      (vector-ref (vector-ref board row) col)
      'out))

;; space?, solid?, door?, room? : cell -> boolean
(define (space? cell) (eq? cell 'space))
(define (solid? cell) (or (eq? cell 'out) (room? cell)))
(define (door? cell) (memq cell '(c b l s o h k d u)))
(define (room? cell) (memq cell '(C B L S O H K D U)))

;; door->room : door -> room
(define (door->room door)
  (hash-ref #hasheq((c . C) (b . B) (l . L)
                            (s . S) (o . O) (h . H)
                            (k . K) (d . D) (u . U))
            door))

;; find-doors : symbol -> hash-of-pos
;;  Values in hash are all #t
(define (find-doors room)
  (for*/fold ([dests #hash()]) ([row (in-range 25)]
                                [col (in-range 25)])
    (let ([c (board-ref row col)])
      (if (and (door? c)
               (eq? room (door->room c)))
          (hash-set dests (list row col) #t)
          dests))))

(test (find-doors 'C) #hash(((5 5) . #t)))
(test (find-doors 'L) #hash(((3 13) . #t) ((7 16) . #t)))
