#lang scheme

(require "board.ss"
         "test.ss")

;; find-roll-dests : dimen dimen nat list-of-pos -> hash
;;  Keys of resulting hash are `pos'es (and vals are #t)
(define (find-roll-dests row col roll occupied)
  
  (define (try dr dc so-far)
    (let ([row (+ row dr)]
          [col (+ col dc)])
      (if (and (<= 0 row 24)
               (<= 0 col 24))
          (let ([c (board-ref row col)])
            (cond
              [(solid? c) so-far]
              [(and (= roll 1)
                    (member (list row col) occupied))
               so-far]
              [else 
               (union
                (if (and (door? c)
                         (not (member (list row col) occupied)))
                    ;; Can enter room
                    (hash-set #hash() (list (door->room c) 'V) #t)
                    #hash())
                (union
                 (find-roll-dests row col (sub1 roll) occupied)
                 so-far))]))
          so-far)))
  
  (cond
    [(zero? roll) (hash-set #hash() (list row col) #t)]
    [else
     (try 0 1 (try 1 0 (try -1 0 (try 0 -1 #hash()))))]))

;; union : hash hash -> hash
;;  Assumes that all values are #t, keys are compared with equal?
(define (union a b)
  (cond
    [((hash-count a) . > . (hash-count b)) (union b a)]
    [else (for/fold ([b b])
            ([k (in-hash-keys a)])
            (hash-set b k #t))]))

(test (union #hash() #hash()) #hash())
(test (union #hash((a . #t)) #hash((b . #t))) #hash((a . #t) (b . #t)))
(test (union #hash((a . #t)) #hash((a . #t))) #hash((a . #t)))
(test (union #hash((a . #t) (b . #t)) #hash((a . #t)))
      #hash((a . #t) (b . #t)))
(test (union #hash((a . #t) (b . #t)) #hash((c . #t))) 
      #hash((a . #t) (b . #t) (c . #t)))

(test (find-roll-dests 7 11 1 null)
      #hash(((6 11) . #t)
            ((8 11) . #t)
            ((7 10) . #t)
            ((7 12) . #t)))
(test (find-roll-dests 7 11 2 null)
      #hash(((6 10) . #t)
            ((6 12) . #t)
            ((7 11) . #t)
            ((8 10) . #t)
            ((8 12) . #t)
            ((7 9) . #t)
            ((7 13) . #t)))
(test (find-roll-dests 7 9 1 null)
      #hash(((6 9) . #t)
            ((7 8) . #t)
            ((7 10) . #t)
            ((8 9) . #t)
            ((B V) . #t)))
(test (find-roll-dests 7 9 2 null)
      #hash(((7 9) . #t)
            ((6 8) . #t)
            ((6 10) . #t)
            ((8 8) . #t)
            ((7 7) . #t)
            ((7 11) . #t)
            ((9 9) . #t)
            ((8 8) . #t)
            ((8 10) . #t)
            ((B V) . #t)))
(test (find-roll-dests 7 9 2 '((6 9)))
      #hash(((7 9) . #t)
            ((6 8) . #t)
            ((6 10) . #t)
            ((8 8) . #t)
            ((7 7) . #t)
            ((7 11) . #t)
            ((9 9) . #t)
            ((8 8) . #t)
            ((8 10) . #t)))

(for* ([row (in-range 25)]
       [col (in-range 25)])
  (time (find-roll-dests row col 12 null)))
